#!/usr/bin/python
# -*- coding: utf-8 -*-
'''
Default config variables which maybe overridden by a user config.
'''
import os.path as osp

# Lotka-Volterra Parameters
LV_PARAMS = {
'A' : 10.0,
'B' : 7.0,
'C' : 3.0,
'D' : 5.0,
'STEP_SIZE' : 0.01,
'INITIAL_TIME' : 0,
'INITIAL_PREY_POPULATION' : 60,
'INITIAL_PREDATOR_POPULATION' : 25,
'MAX_ITERATIONS' : 200
}

# PATHS
PROJECT_PATH = osp.abspath(osp.join(osp.dirname(__file__), '..'))
RESULTS_DIR = osp.join(PROJECT_PATH, 'results')



#!/usr/bin/python
# -*- coding: utf-8 -*-
'''
Logs experimental configuration in the log file.
'''
import logging
import causal_inference.config as CF

def log_separator_line():
    '''
    Adds a line to log for easy readability.
    '''
    logging.info('-----------------------------------------------')

def log_LV_params():
    '''
    Logs Lotka-Volterra parameters to logfile.
    '''
    log_separator_line()
    logging.info('Lotka-Volterra parameters:')
    logging.info('A: %d', CF.LV_PARAMS['A'])
    logging.info('B: %d', CF.LV_PARAMS['B'])
    logging.info('C: %d', CF.LV_PARAMS['C'])
    logging.info('D: %d', CF.LV_PARAMS['D'])
    logging.info('INITIAL_PREY_POPULATION: %d', CF.LV_PARAMS['INITIAL_PREY_POPULATION'])
    logging.info('INITIAL_PREDATOR_POPULATION: %d', CF.LV_PARAMS['INITIAL_PREDATOR_POPULATION'])
    logging.info('STEP_SIZE: %f', CF.LV_PARAMS['STEP_SIZE'])
    log_separator_line()



#!/usr/bin/env python3
# flake8: noqa
from causal_inference import LotkaVolterra
from causal_inference import LV_PARAMS

def test_lotka_volterra():
    model = LotkaVolterra()

    assert model.A == LV_PARAMS['A']
    assert model.B == LV_PARAMS['B']
    assert model.C == LV_PARAMS['C']
    assert model.D == LV_PARAMS['D']



#!/usr/bin/env python3
# flake8: noqa
'''
Setup script for the repository.
'''
import os
import codecs
from setuptools import find_packages, setup, Distribution

def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return codecs.open(file_path, encoding='utf-8').read()

requirements = []
with open('requirements.txt') as f:
    for line in f:
        stripped = line.split("#")[0].strip()
        if len(stripped) > 0:
            requirements.append(stripped)

# Currently incompatible with PyPi
# use_scm = {"write_to": "causal_inference/_version.py"}

def local_scheme(version):
    return ""

setup(
    name = 'causal-inference-population-dynamics',
    author='Pranjal Dhole',
    author_email='dhole.pranjal@gmail.com',
    license='MIT',
    url='https://github.com/artificial-life-lab/population-dynamics',
    description='Library to conduct experiments in population dynamics.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    distclass = Distribution,
    packages = find_packages(),
    python_requires='>=3.7',
    install_requires=requirements,
    use_scm_version={"local_scheme": local_scheme},
    setup_requires=['setuptools_scm'],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-

import argparse
import os
import logging
import datetime
import h5py
import matplotlib.pyplot as plt

from causal_inference.config import LV_PARAMS, RESULTS_DIR
from causal_inference.utils.log_config import log_LV_params


class LotkaVolterra():
    '''
    Class simulates predator-prey dynamics and solves it with 4th order Runge-Kutta method.
    '''
    def __init__(self):
        self.A = LV_PARAMS['A']
        self.B = LV_PARAMS['B']
        self.C = LV_PARAMS['C']
        self.D = LV_PARAMS['D']
        self.time = LV_PARAMS['INITIAL_TIME']
        self.step_size = LV_PARAMS['STEP_SIZE']
        self.max_iterations = LV_PARAMS['MAX_ITERATIONS']

        self.prey_population = LV_PARAMS['INITIAL_PREY_POPULATION']
        self.predator_population = LV_PARAMS['INITIAL_PREDATOR_POPULATION']

        self.time_stamp = [self.time]
        self.prey_list = [self.prey_population]
        self.predator_list = [self.predator_population]
    
    def compute_prey_rate(self, current_prey, current_predators):
        return self.A * current_prey - self.B * current_prey * current_predators

    def compute_predator_rate(self, current_prey, current_predators):
        return - self.C * current_predators + self.D * current_prey * current_predators
    
    def runge_kutta_update(self, current_prey, current_predators):
        self.time = self.time + self.step_size
        self.time_stamp.append(self.time)

        k1_prey = self.step_size * self.compute_prey_rate(current_prey, current_predators)
        k1_pred = self.step_size * self.compute_predator_rate(current_prey, current_predators)

        k2_prey = self.step_size * self.compute_prey_rate(current_prey + 0.5 * k1_prey, current_predators + 0.5 * k1_pred)
        k2_pred = self.step_size * self.compute_predator_rate(current_prey + 0.5 * k1_prey, current_predators + 0.5 * k1_pred)

        k3_prey = self.step_size * self.compute_prey_rate(current_prey + 0.5 * k2_prey, current_predators + 0.5 * k2_pred)
        k3_pred = self.step_size * self.compute_predator_rate(current_prey + 0.5 * k2_prey, current_predators + 0.5 * k2_pred)

        k4_prey = self.step_size * self.compute_prey_rate(current_prey + k3_prey, current_predators + k3_pred)
        k4_pred = self.step_size * self.compute_predator_rate(current_prey + k3_prey, current_predators + k3_pred)

        new_prey_population = current_prey + 1/6 * (k1_prey + 2 * k2_prey + 2 * k3_prey + k4_prey)
        new_predator_population = current_predators + 1/6 * (k1_pred + 2 * k2_pred + 2 * k3_pred + k4_pred)

        self.prey_list.append(new_prey_population)
        self.predator_list.append(new_predator_population)

        return new_prey_population, new_predator_population

    def _save_population(self):
        filename = os.path.join(RESULTS_DIR, 'populations.h5')
        hf = h5py.File(filename, 'w')
        hf.create_dataset('prey_pop', data=self.predator_list)
        hf.create_dataset('pred_pop', data=self.predator_list)
        hf.close()

    def _solve(self):
        current_prey, current_predators = self.prey_population, self.predator_population
        logging.info('Computing population over time...')
        for gen_idx in range(self.max_iterations):
            current_prey, current_predators = self.runge_kutta_update(current_prey, current_predators)
            logging.info('Gen: %d | Prey population: %d | Predator population: %d', gen_idx, current_prey, current_predators)
        print('Done!')
    
    def plot_population_over_time(self, save=True, filename='predator_prey'):
        _ = plt.figure(figsize=(15, 5))
        PreyLine, = plt.plot(self.prey_list , color='g')
        PredatorsLine, = plt.plot(self.predator_list, color='r')
        plt.legend([PreyLine, PredatorsLine], ['Prey', 'Predators'])
        plt.ylabel('Population')
        plt.xlabel('Time')
        if save:
            plt.savefig(os.path.join(RESULTS_DIR, f"{filename}.svg"),
                        format='svg', transparent=False, bbox_inches='tight')
        else:
            plt.show()
        plt.close()

def main():
    logging.info('Lotka-Volterra predator-prey dynamics with 4th order Runge-Kutta method')
    m = LotkaVolterra()
    log_LV_params()
    m._solve()
    m._save_population()
    m.plot_population_over_time()

if __name__ == '__main__':
    PARSER = argparse.ArgumentParser()
    PARSER.add_argument('-log', '--logfile', help='name of the logfile', default='log')
    PARSER.add_argument('-out', '--outdir', help='Where to place the results', default='lv_simulation')
    ARGS = PARSER.parse_args()

    RESULTS_DIR = os.path.join(RESULTS_DIR, '{}_{}'.format(datetime.datetime.now().strftime("%Y%h%d_%H_%M_%S"), str(ARGS.outdir)))

    if not os.path.exists(RESULTS_DIR):
        os.makedirs(RESULTS_DIR)

    LOG_FILE = os.path.join(RESULTS_DIR, f"{ARGS.logfile}.txt")     # write logg to this file
    logging.basicConfig(
        level=logging.INFO,
        handlers=[
            logging.FileHandler(LOG_FILE),
            logging.StreamHandler()
        ]
    )

    main()
